/*
 * 17:37:42 14/05/00
 *
 * CaffeineClient.java - The Caffeine Client
 * Copyright (C) 1999 Romain Guy
 * powerteam@chez.com
 * http://www.chez.com/powerteam
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.chez.powerteam.caffeine;

import java.io.*;
import java.net.*;
import java.util.*;
import com.chez.powerteam.caffeine.event.*;
import com.chez.powerteam.caffeine.filter.*;

/**
 * The <code>CaffeineClient</code> is a client which can send
 * and receive datas from a caffeine server. The client uses
 * a filter to send objects and receive objects.
 * @author Romain Guy
 */

public class CaffeineClient implements Runnable
{
  // the filter used to turn objects into strings and reverse
  private CaffeineFilter filter;
  // the list of list filters
  private Vector listFilters = new Vector();
  // this array contains all the listeners linked to this class
  public Vector listeners = new Vector();
  // the socket which stand for the server
  private Socket server;
  // the stream used to send datas to the server
  private PrintWriter writer;
  // the stream used to read datas received from the server
  private BufferedReader reader;
  // the thread
  private Thread network;

  /**
   * Instanciates a new client, open a connection to the server and
   * open the associated streams. These streams will be used to read and
   * write from/to the server.
   * @param ip The server IP
   * @param port The port of te server opened for the client
   */

  public CaffeineClient(String ip, int port)
  {
    try
    {
      server = new Socket(ip, port);

      reader = new BufferedReader(new InputStreamReader(server.getInputStream()));
      writer = new PrintWriter(new OutputStreamWriter(server.getOutputStream()), true);
    } catch (IOException ioe) {
      System.err.println("[Server is not running]");
      System.exit(2);
    }

    InetAddress url = server.getInetAddress();
    System.out.println("Server IP: " + url.getHostAddress() + "; Server name: " +
                       url.getHostName());

    setFilter(new DefaultCaffeineFilter(this));

    network = new Thread(this);
    network.setName("Caffeine Client: Logged on: " + url.getHostAddress());
    network.setPriority(Thread.NORM_PRIORITY);
    network.start();
  }

  /**
   * Forces the thread to stop and close server.
   */

  public void stop()
  {
    network = null;
    listFilters.removeAllElements();

    writer.println("/kill");
    writer.flush();

    // we wait .5 second to ensure our '/kill' message has
    // been sent. Otherwise, Java encounters socket problem
    try
    {
      Thread.sleep(500);
    } catch (InterruptedException ie) { }    

    try
    {
      server.close();
    } catch (IOException ioe) { }
    fireEvent("CAFFEINE_CONNECTION_CLOSED");
  }

  /**
   * Get strings from server and filter them.
   */

  public void run()
  {
    String line;
    while (network != null)
    {
      try
      {
        if ((line = reader.readLine()) != null)
        {
          if (line.equals("/kill"))
          {
            System.out.println("[Deconnection]");
            stop();
          } else {
            if (listFilters != null && listFilters.size() > 0)
            {
              // we send the string to the filters list reversed
              for (int i = listFilters.size() - 1; i >= 0; i--)
                line = ((CaffeineListFilter) listFilters.elementAt(i)).filterStringIn(line);
            }
            filter.filterIn(line);
          }
        }
      } catch (IOException ioe) {
        System.err.println("[Client error]\n" + ioe);
        System.exit(2);
      }
      network.yield();
    }
  }

  /**
   * Removes all the listeners associated with this client.
   */

  public void removeAllListeners()
  {
    listeners.removeAllElements();
  }

  /**
   * Adds a network listener to this class.
   * @param l The <code>NetworkListener</code> to add
   */

  public void addNetworkListener(NetworkListener l)
  {
    listeners.addElement(l);
  }

  /**
   * Remove a specified network listener from the list.
   * @param l The listener to remove
   */

  public void removeNetworkListener(NetworkListener l)
  {
    listeners.removeElement(l);
  }


  /**
   * Removes all the list filters associated with this client.
   */

  public void removeAllListFilters()
  {
    listFilters.clear();
  }

  /**
   * Adds a list filter in the chain.
   * @param l The <code>CaffeineListFilter</code> to add
   */

  public void addListFilter(CaffeineListFilter l)
  {
    listFilters.addElement(l);
  }

  /**
   * Remove a specified list filter from the list.
   * @param l The filter to remove
   */

  public void removeListFilter(CaffeineListFilter l)
  {
    listFilters.removeElement(l);
  }

  /**
   * Send an object to the server. Other applications
   * should use this method to get advantage of the filter.
   * @param o The object to be sent
   */

  public void send(Object o)
  {
    filter.filterOut(o);
  }

  /**
   * Send a network event to the listeners. This happens when
   * we receive a string from the server through network.
   * @param o The object to be embedded into the event
   */

  public void fireEvent(Object o)
  {
    NetworkEvent e = new NetworkEvent(o);
    // we send the event to all the listeners available
    for (int i = 0; i < listeners.size(); i++)
      ((NetworkListener) listeners.elementAt(i)).networkEvent(e);
  }

  /**
   * Send a string to the server. This is called when user
   * application needs to transfer informations to other
   * clients.
   * @param s The <code>String</code> to send
   */

  public void sendString(String s)
  {
    // paranoia
    if (writer != null)
    {
      if (listFilters != null && listFilters.size() > 0)
      {
        for (int i = 0; i < listFilters.size(); i++)
          s = ((CaffeineListFilter) listFilters.elementAt(i)).filterStringOut(s);
      }
      writer.println(s);
    }
  }

  /**
   * As <code>CaffeineFilter</code> is an interface, you may need to
   * determine which kind of filter is being used. Some coffees sometimes
   * need specials filters ;-)
   * @return Current used filter
   */

  public CaffeineFilter getFilter()
  {
    return filter;
  }

  /**
   * Set the new filter to be used instead of <code>DefaultCaffeineFilter</code>.
   * Call it when you want to do more than <code>toString()</code> on sending
   * and nothing on receiving.
   * @param filter The new <code>CaffeineFilter</code> to be used
   */

  public void setFilter(CaffeineFilter filter)
  {
    this.filter = filter;
  }

  /**
   * The advanced caffeine filter needs to use the writer logged to
   * the caffeine server.
   * @return The <code>OutputStream</code> used to send datas to the server
   */

  public OutputStream getOutputStream() throws IOException
  {
    return server.getOutputStream();
  }

  /**
   * The advanced caffeine filter needs to use the reader logged to
   * the caffeine server.
   * @return The <code>InputStream</code> used to receive datas from the server
   */

  public InputStream getInputStream() throws IOException
  {
    return server.getInputStream();
  }
}

// End of CaffeineClient.java
