/*
 * 20:00:09 21/04/00
 *
 * CaffeineServer.java - The Caffeine server
 * Copyright (C) 1999-2000 Romain Guy
 * powerteam@chez.com
 * http://www.chez.com/powerteam
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.chez.powerteam.caffeine;

import java.io.*;
import java.net.*;
import java.util.*;
import com.chez.powerteam.caffeine.event.*;
import com.chez.powerteam.caffeine.exception.*;

/**
 * The <code>CaffeineServer</code> is a server which can accept a
 * fixed number of clients (which can be set).
 * @author Romain Guy
 */

public class CaffeineServer implements Runnable
{
  /** The default maximum number of connections allowed */
  public static final int DEFAULT_MAX_CLIENTS = 4;

  private String password;
  private volatile boolean alive = true;
  // the port used by the server
  private int serverPort;
  // an ID used to identify the server
  private String serverName;
  // set maximum connections allowed
  private int maxClients = DEFAULT_MAX_CLIENTS;
  // the array of clients
  private Vector clients;
  // this is the server itself
  private ServerSocket server;
  // the clients seeker
  private Thread seeker;
  // this array contains all the listeners linked to this class
  public Vector listeners = new Vector();

  /**
   * Create a new running server.
   * @param serverPort The ID of the port to be used by the server
   */

  public CaffeineServer(int serverPort, String password)
  {
    this(serverPort, new Date().toString(), password);
  }

  /**
   * Create a new running server.
   * @param serverPort The ID of the port to be used by the server
   * @param serverName A <code>String</code> used to identify the server
   * @param password The password to log in
   */

  public CaffeineServer(int serverPort, String serverName, String password)
  {
    this(serverPort, serverName, password, DEFAULT_MAX_CLIENTS);
  }

  /**
   * Create a new running server.
   * @param serverPort The ID of the port to be used by the server
   * @param serverName A <code>String</code> used to identify the server
   * @param password The password to log in
   * @param maxClients The maximum number of clients allowed (4 by default)
   */

  public CaffeineServer(int serverPort, String serverName, String password, int maxClients)
  {
    // we save the settings
    this.serverPort = serverPort;
    this.serverName = serverName;
    this.password = password;
    this.maxClients = maxClients;

    // if I choosed a Vector, it is just to be able to
    // increase its size (more difficult with a simple array)
    clients = new Vector(this.maxClients);

    // we start the server
    startServer();

    // paranoia
    if (server == null)
    {
      System.err.println("[Server is not running]");
      System.exit(1);
    }
  }

  /**
   * Start the server.
   */

   private void startServer()
   {
     try
     {
       server = new ServerSocket(serverPort);
       server.setSoTimeout(1000);
     } catch (IOException ioe) {
       System.err.println("[Cannot initialize Server]\n" + ioe);
       System.exit(1);
     }

     // we display IP address and port number to ensure
     // the server is ours ;-)
     try
     {
       InetAddress url = InetAddress.getLocalHost();
       System.out.println("[Server is running]\nServer IP: " + url.getHostAddress() +
                          "; Server name: " + url.getHostName() + "; Server port: "
                          + serverPort);
     } catch (UnknownHostException uhe) { }

     startSeeker();
   }

  /**
   * This resolves clients requests. If we have enough clients,
   * we stop the thread.
   */

  public void run()
  {
    while (alive)
      getClient();
  }

  /**
   * This method get a Client from a ServerSocket.
   * @return true if we managed to get a client
   */

  private void getClient()
  {
    Socket client;
    try
    {
      // wait for a client to connect on our port
      client = server.accept();
      if (clients.size() == getMaxClients())
      {
        System.out.println("[Server cannot handle more clients]");
        (new PrintWriter(new OutputStreamWriter(client.getOutputStream()))).println("/kill");
        return;
      }

      new Authorizer(client, password);
      // addClient(new CaffeineSocket(this, client, password));
    } catch (IOException ioe) {
      if (seeker == null)
        alive = false;
      return;
    }
  }

  /**
   * Stop the seeker by killing the thread, close
   * the server socket, free memory and exit.
   */

   public void stop()
   {
     stop(false);
   }

  /**
   * Stop the seeker by killing the thread, close
   * the server socket, free memory and exit.
   * @param killJMV If true, we <code>System.exit(0);</code>
   */

  public void stop(boolean killJVM)
  {
    // kill thread
    stopSeeker();

    // kill clients
    for (int i = 0; i < clients.size(); i++)
    {
      CaffeineSocket kill = (CaffeineSocket) clients.elementAt(i);
      kill.getOut().println("/kill");
      kill.close();
    }
    clients.removeAllElements();

    // close server
    try
    {
      server.close();
    } catch (IOException ioe) { }

    if (killJVM)
      System.exit(0);
  }

  /**
   * Stop the seeker by killing the thread.
   */

  public void stopSeeker()
  {
    seeker = null;
  }

  /**
   * Start the clients seeker.
   */

  public void startSeeker()
  {
    seeker = new Thread(this);
    seeker.setName("CaffeineServer: " + serverName + ": Client Seeker");
    seeker.setPriority(Thread.MIN_PRIORITY);
    seeker.start();
  }

  /**
   * Removes all the listeners associated with this class.
   */

  public void removeAllListeners()
  {
    listeners.removeAllElements();
  }

  /**
   * Adds a server listener to this class.
   * @param l The <code>NetworkListener</code> to add
   */

  public void addServerListener(ServerListener l)
  {
    listeners.addElement(l);
  }

  /**
   * Remove a specified server listener from the list.
   * @param l The listener to remove
   */

  public void removeServerListener(ServerListener l)
  {
    listeners.removeElement(l);
  }

  /**
   * Send a network event to the listeners. This happens when
   * we receive a string from the server through network.
   * @param o The object to be embedded into the event
   */

  public void fireEvent(String line, CaffeineSocket socket)
  {
    ServerEvent e = new ServerEvent(line, socket);
    // we send the event to all the listeners available
    for (int i = 0; i < listeners.size(); i++)
      ((ServerListener) listeners.elementAt(i)).serverEvent(e);
  }

  /**
   * If the client exits, we kill it.
   * @param IP The IP of the client to be killed
   */

  public synchronized void killClient(String IP)
  {
    for (int i = 0; i < clients.size(); i++)
    {
      CaffeineSocket kill = (CaffeineSocket) clients.elementAt(i);
      if (kill.getIP().equals(IP))
      {
        kill.getOut().println("/kill");
        kill.close();
        clients.removeElement(kill);
        return;
      }
    }
  }

  /**
   * Add a client to the list only if we haven't reached the limit.
   * @param client A <code>CaffeineSocket</code> refering to a client
   */

  public synchronized void addClient(CaffeineSocket client)
  {
    if (clients.size() < maxClients)
      clients.addElement(client);
  }

  /**
   * Another class may need to handle the clients, too.
   * @return An array of <code>Socket</code>, one per client logged
   */

  public synchronized CaffeineSocket[] getClients()
  {
    CaffeineSocket[] clientsSockets = new CaffeineSocket[clients.size()];
    for (int i = 0; i < clientsSockets.length; i++)
      clientsSockets[i] = (CaffeineSocket) clients.elementAt(i);
    return clientsSockets;
  }

  /**
   * An external class, which could be handling the server, may need
   * to know the amount of clients.
   * @return The amount of clients handled by the server
   */

  public synchronized int clientsCount()
  {
    return clients.size();
  }

  /**
   * If a class needs to identify a server, it can
   * use its name.
   * @return The name of the server
   */

  public String getName()
  {
    return serverName;
  }

  /**
   * Set the name of the server.
   * @param name The new name of the server
   */

  public void setName(String name)
  {
    serverName = name;
  }

  /**
   * Knowing the running port is very useful for exceptions.
   * @return The port number used by the server
   */

  public int getPort()
  {
    return serverPort;
  }

  /**
   * The server has a limited number of clients.
   * @return The maximum number of clients allowed
   */

  public synchronized int getMaxClients()
  {
    return maxClients;
  }

  /**
   * Set the maximum number of available connections.
   * @param max The maximum number connections the server can handle
   * @throws A <code>ServerException</code> if <code>max</code> is &lt; 1
   */

  public synchronized void setMaxClients(int max) throws CaffeineServerException
  {
    if (max < 1)
      throw new CaffeineServerException(this,
                "Attempted to set maximum connections under 0.");
    else if (max < maxClients)
    {
      for (int i = max; i < clients.size(); i++)
      {
        CaffeineSocket kill = (CaffeineSocket) clients.elementAt(i);
        kill.getOut().println("/kill");
        kill.close();
        clients.removeElement(kill);
      }
    }

    maxClients = max;
    clients.ensureCapacity(max);
  }

  /**
   * <code>Authorize</code> is a thread which checks password given by client.
   * If the password is wrong, the connection is refused.
   */

  class Authorizer extends Thread
  {
    private Socket client;
    private String password;

    Authorizer(Socket client, String password)
    {
      super();

      this.client = client;
      this.password = password;

      setPriority(MIN_PRIORITY);
      this.setName("CaffeineServer: Authorizer");
      start();
    }

    public void run()
    {
      if (client == null)
        return;

      if (password != null)
      {
        try
        {
          InputStream in = client.getInputStream();
          OutputStream out = client.getOutputStream();
          if (in == null || out == null)
            return;

          BufferedReader reader = new BufferedReader(new InputStreamReader(in));
          PrintWriter writer = new PrintWriter(new OutputStreamWriter(out), true);

          writer.println("Welcome to Caffeine. Enter your password:");
          String _password = reader.readLine();
          if (!_password.equals(password))
          {
            writer.println("\nWrong password !");          
            client.close();
            return;
          }

          writer.println("\nYou are now logged on Caffeine server " + CaffeineServer.this.getName());
          writer.print("\n> ");
          writer.flush();

        } catch (NullPointerException npe) {
          return;
        } catch (IOException ioe) {
          return;
        }
      }

      try
      {
        addClient(new CaffeineSocket(CaffeineServer.this, client));
      } catch (CaffeineSocketException cse) {
        System.out.println(cse.toString());
        return;
      }

      System.out.println("[Get client]");
      InetAddress url = client.getInetAddress();
      System.out.println("Client IP: " + url.getHostAddress() + "; Client name: " +
                         url.getHostName());
    }
  }
}

// End of CaffeineServer.java
