/*
 * CaffeineSocket.java - An Object Socket for Caffeine
 * Copyright (C) 1999 Romain Guy
 * powerteam@chez.com
 * http://www.chez.com/powerteam
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.chez.powerteam.caffeine;

import java.io.*;
import java.net.*;
import com.chez.powerteam.caffeine.exception.*;

/**
 * The CaffeineSocket acts as a normal socket but provides
 * more functions. For instance, you can directly get the
 * writer and the reader. Moreover, this socket dispatches
 * the String sent through it to server to the other clients
 * @author Romain Guy
 */

public class CaffeineSocket implements Runnable
{
  // to send String to the client
  private PrintWriter writer;
  // to receive String from the client
  private BufferedReader reader;
  // infos about the client
  private InetAddress client;
  // the socket
  private Socket socket;
  // the dispatching thread
  private Thread sender;
  // parent
  private CaffeineServer parent;

  /**
   * Create a new CaffeineSocket, constructed from a standard Socket.
   * @param socket The <code>Socket</code> used to construct the <code>CaffeineSocket</code>
   */

  public CaffeineSocket(CaffeineServer parent, Socket socket) throws CaffeineSocketException
  {
    this.socket = socket;
    this.parent = parent;

    try
    {
      reader = new BufferedReader(new InputStreamReader(socket.getInputStream()));
      writer = new PrintWriter(new OutputStreamWriter(socket.getOutputStream()), true);
    } catch (IOException ioe) {
      throw new CaffeineSocketException(socket, "Cannot open streams to/from client.");
    }

    client = socket.getInetAddress();

    // the thread which handles the dispatching
    sender = new Thread(this);
    sender.setName("CaffeineSocket: " + getIP() + ": Client: " + getName());
    sender.setPriority(Thread.NORM_PRIORITY);
    sender.start();
  }

  /**
   * Each CaffeineSocket manages its messages and send'em
   * to other CaffeineSockets.
   */

  public void run()
  {
    String request;
//    CaffeineSocket cclient;
//    CaffeineSocket clients[];

    while (sender != null)
    {
      try
      {
        if ((request = reader.readLine()) != null)
        {
          parent.fireEvent(request, this);
          if (request.equals("/kill"))
          {
            parent.killClient(getIP());
            return;
          }
////////////////////////////////////////
// DEBUT MODIFICATION
////////////////////////////////////////
// retir car nous ne voulons PAS que
// les autres clients recoivent ce qui
// est reu par le serveur
////////////////////////////////////////
/*
          clients = parent.getClients();
          for (int i = 0; i < clients.length; i++)
          {
            cclient = clients[i];
            // if we synchronize before testing the socket,
            // we could encounter the case the class tries
            // to synchronize with itself
            if (cclient != this)
            {
              synchronized(cclient)
              {
                cclient.getOut().println(request);
              }
            }
          }
*/
////////////////////////////////////////
// FIN MODIFICATION
////////////////////////////////////////
        }
      } catch (Exception ioe) {
        if (sender != null)
        {
          // parent.stop();
          parent.killClient(getIP());
          return;
        }
      }
      Thread.yield();
    }
  }

  /**
   * Return the socket used.
   */

   public Socket getSocket()
   {
     return socket;
   }

  /**
   * Required to receive String from the client.
   * @return A <code>BufferedReader</code>
   */

  public BufferedReader getIn()
  {
    return reader;
  }

  /**
   * Required to send String to the client.
   * @return A <code>PrintWriter</code>
   */

  public PrintWriter getOut()
  {
    return writer;
  }

  /**
   * Needed to compare clients.
   * @return The client's IP
   */

  public String getIP()
  {
    return client.getHostAddress();
  }

  /**
   * Needed to compare clients.
   * @return The client's computer name
   */

  public String getName()
  {
    return client.getHostName();
  }

  /**
   * Kill the socket and close its streams.
   */

  public void close()
  {
    sender = null;
    try
    {
      //reader.close();
      //writer.flush();
      //writer.close();
      socket.close();
    } catch (IOException ioe) { }
  }
}

// End of CaffeineSocket.java
